@@ -1,95 +1,69 @@
-import warnings
 from encodings.aliases import aliases
 from hashlib import sha256
 from json import dumps
-from typing import Optional, List, Tuple, Set
-from collections import Counter
-from re import sub, compile as re_compile
+from typing import Any, Dict, Iterator, List, Optional, Tuple, Union
 
-from charset_normalizer.constant import TOO_BIG_SEQUENCE
-from charset_normalizer.md import mess_ratio
-from charset_normalizer.utils import iana_name, is_multi_byte_encoding, unicode_range
+from .constant import TOO_BIG_SEQUENCE
+from .utils import iana_name, is_multi_byte_encoding, unicode_range
 
 
 class CharsetMatch:
     def __init__(
-            self,
-            payload: bytes,
-            guessed_encoding: str,
-            mean_mess_ratio: float,
-            has_sig_or_bom: bool,
-            languages: "CoherenceMatches",
-            decoded_payload: Optional[str] = None
+        self,
+        payload: bytes,
+        guessed_encoding: str,
+        mean_mess_ratio: float,
+        has_sig_or_bom: bool,
+        languages: "CoherenceMatches",
+        decoded_payload: Optional[str] = None,
     ):
-        self._payload = payload  # type: bytes
+        self._payload: bytes = payload
 
-        self._encoding = guessed_encoding  # type: str
-        self._mean_mess_ratio = mean_mess_ratio  # type: float
-        self._languages = languages  # type: CoherenceMatches
-        self._has_sig_or_bom = has_sig_or_bom  # type: bool
-        self._unicode_ranges = None  # type: Optional[List[str]]
+        self._encoding: str = guessed_encoding
+        self._mean_mess_ratio: float = mean_mess_ratio
+        self._languages: CoherenceMatches = languages
+        self._has_sig_or_bom: bool = has_sig_or_bom
+        self._unicode_ranges: Optional[List[str]] = None
 
-        self._leaves = []  # type: List[CharsetMatch]
-        self._mean_coherence_ratio = 0.  # type: float
+        self._leaves: List[CharsetMatch] = []
+        self._mean_coherence_ratio: float = 0.0
 
-        self._output_payload = None  # type: Optional[bytes]
-        self._output_encoding = None  # type: Optional[str]
+        self._output_payload: Optional[bytes] = None
+        self._output_encoding: Optional[str] = None
 
-        self._string = decoded_payload  # type: Optional[str]
+        self._string: Optional[str] = decoded_payload
 
-    def __eq__(self, other) -> bool:
+    def __eq__(self, other: object) -> bool:
         if not isinstance(other, CharsetMatch):
-            raise TypeError('__eq__ cannot be invoked on {} and {}.'.format(str(other.__class__), str(self.__class__)))
+            raise TypeError(
+                "__eq__ cannot be invoked on {} and {}.".format(
+                    str(other.__class__), str(self.__class__)
+                )
+            )
         return self.encoding == other.encoding and self.fingerprint == other.fingerprint
 
-    def __lt__(self, other) -> bool:
+    def __lt__(self, other: object) -> bool:
         """
         Implemented to make sorted available upon CharsetMatches items.
         """
         if not isinstance(other, CharsetMatch):
             raise ValueError
 
-        chaos_difference = abs(self.chaos - other.chaos)  # type: float
+        chaos_difference: float = abs(self.chaos - other.chaos)
+        coherence_difference: float = abs(self.coherence - other.coherence)
 
-        # Bellow 1% difference --> Use Coherence
-        if chaos_difference < 0.01:
+        # Below 1% difference --> Use Coherence
+        if chaos_difference < 0.01 and coherence_difference > 0.02:
             return self.coherence > other.coherence
+        elif chaos_difference < 0.01 and coherence_difference <= 0.02:
+            # When having a difficult decision, use the result that decoded as many multi-byte as possible.
+            return self.multi_byte_usage > other.multi_byte_usage
 
         return self.chaos < other.chaos
 
     @property
-    def chaos_secondary_pass(self) -> float:
-        """
-        Check once again chaos in decoded text, except this time, with full content.
-        Use with caution, this can be very slow.
-        Notice: Will be removed in 3.0
-        """
-        warnings.warn("chaos_secondary_pass is deprecated and will be removed in 3.0", DeprecationWarning)
-        return mess_ratio(
-            str(self),
-            1.
-        )
-
-    @property
-    def coherence_non_latin(self) -> float:
-        """
-        Coherence ratio on the first non-latin language detected if ANY.
-        Notice: Will be removed in 3.0
-        """
-        warnings.warn("coherence_non_latin is deprecated and will be removed in 3.0", DeprecationWarning)
-        return 0.
-
-    @property
-    def w_counter(self) -> Counter:
-        """
-        Word counter instance on decoded text.
-        Notice: Will be removed in 3.0
-        """
-        warnings.warn("w_counter is deprecated and will be removed in 3.0", DeprecationWarning)
-        not_printable_pattern = re_compile(r'[0-9\W\n\r\t]+')
-        string_printable_only = sub(not_printable_pattern, ' ', str(self).lower())
-
-        return Counter(string_printable_only.split())
+    def multi_byte_usage(self) -> float:
+        return 1.0 - (len(str(self)) / len(self.raw))
 
     def __str__(self) -> str:
         # Lazy Str Loading
@@ -102,7 +76,11 @@ class CharsetMatch:
 
     def add_submatch(self, other: "CharsetMatch") -> None:
         if not isinstance(other, CharsetMatch) or other == self:
-            raise ValueError("Unable to add instance <{}> as a submatch of a CharsetMatch".format(other.__class__))
+            raise ValueError(
+                "Unable to add instance <{}> as a submatch of a CharsetMatch".format(
+                    other.__class__
+                )
+            )
 
         other._string = None  # Unload RAM usage; dirty trick.
         self._leaves.append(other)
@@ -116,7 +94,7 @@ class CharsetMatch:
         """
         Encoding name are known by many name, using this could help when searching for IBM855 when it's listed as CP855.
         """
-        also_known_as = []  # type: List[str]
+        also_known_as: List[str] = []
         for u, p in aliases.items():
             if self.encoding == u:
                 also_known_as.append(p)
@@ -153,9 +131,13 @@ class CharsetMatch:
                 return "English"
 
             # doing it there to avoid circular import
-            from charset_normalizer.cd import mb_encoding_languages, encoding_languages
+            from charset_normalizer.cd import encoding_languages, mb_encoding_languages
 
-            languages = mb_encoding_languages(self.encoding) if is_multi_byte_encoding(self.encoding) else encoding_languages(self.encoding)
+            languages = (
+                mb_encoding_languages(self.encoding)
+                if is_multi_byte_encoding(self.encoding)
+                else encoding_languages(self.encoding)
+            )
 
             if len(languages) == 0 or "Latin Based" in languages:
                 return "Unknown"
@@ -171,7 +153,7 @@ class CharsetMatch:
     @property
     def coherence(self) -> float:
         if not self._languages:
-            return 0.
+            return 0.0
         return self._languages[0][1]
 
     @property
@@ -201,12 +183,12 @@ class CharsetMatch:
     def alphabets(self) -> List[str]:
         if self._unicode_ranges is not None:
             return self._unicode_ranges
-        detected_ranges = set()  # type: Set[str]
-        for character in str(self):
-            detected_ranges.add(
-                unicode_range(character)
-            )
-        self._unicode_ranges = sorted(list(detected_ranges))
+        # list detected ranges
+        detected_ranges: List[Optional[str]] = [
+            unicode_range(char) for char in str(self)
+        ]
+        # filter and sort
+        self._unicode_ranges = sorted(list({r for r in detected_ranges if r}))
         return self._unicode_ranges
 
     @property
@@ -218,18 +200,6 @@ class CharsetMatch:
         """
         return [self._encoding] + [m.encoding for m in self._leaves]
 
-    def first(self) -> "CharsetMatch":
-        """
-        Kept for BC reasons. Will be removed in 3.0.
-        """
-        return self
-
-    def best(self) -> "CharsetMatch":
-        """
-        Kept for BC reasons. Will be removed in 3.0.
-        """
-        return self
-
     def output(self, encoding: str = "utf_8") -> bytes:
         """
         Method to get re-encoded bytes payload using given target encoding. Default to UTF-8.
@@ -254,14 +224,14 @@ class CharsetMatches:
     Container with every CharsetMatch items ordered by default from most probable to the less one.
     Act like a list(iterable) but does not implements all related methods.
     """
-    def __init__(self, results: List[CharsetMatch] = None):
-        self._results = sorted(results) if results else []  # type: List[CharsetMatch]
 
-    def __iter__(self):
-        for result in self._results:
-            yield result
+    def __init__(self, results: Optional[List[CharsetMatch]] = None):
+        self._results: List[CharsetMatch] = sorted(results) if results else []
 
-    def __getitem__(self, item) -> CharsetMatch:
+    def __iter__(self) -> Iterator[CharsetMatch]:
+        yield from self._results
+
+    def __getitem__(self, item: Union[int, str]) -> CharsetMatch:
         """
         Retrieve a single item either by its position or encoding name (alias may be used here).
         Raise KeyError upon invalid index or encoding not present in results.
@@ -278,17 +248,24 @@ class CharsetMatches:
     def __len__(self) -> int:
         return len(self._results)
 
+    def __bool__(self) -> bool:
+        return len(self._results) > 0
+
     def append(self, item: CharsetMatch) -> None:
         """
         Insert a single match. Will be inserted accordingly to preserve sort.
         Can be inserted as a submatch.
         """
         if not isinstance(item, CharsetMatch):
-            raise ValueError("Cannot append instance '{}' to CharsetMatches".format(str(item.__class__)))
+            raise ValueError(
+                "Cannot append instance '{}' to CharsetMatches".format(
+                    str(item.__class__)
+                )
+            )
         # We should disable the submatch factoring when the input file is too heavy (conserve RAM usage)
         if len(item.raw) <= TOO_BIG_SEQUENCE:
             for match in self._results:
-                if match.fingerprint == item.fingerprint:
+                if match.fingerprint == item.fingerprint and match.chaos == item.chaos:
                     match.add_submatch(item)
                     return
         self._results.append(item)
@@ -314,42 +291,47 @@ CoherenceMatches = List[CoherenceMatch]
 
 
 class CliDetectionResult:
-
-    def __init__(self, path: str, encoding: str, encoding_aliases: List[str], alternative_encodings: List[str], language: str, alphabets: List[str], has_sig_or_bom: bool, chaos: float, coherence: float, unicode_path: Optional[str], is_preferred: bool):
-        self.path = path  # type: str
-        self.unicode_path = unicode_path  # type: Optional[str]
-        self.encoding = encoding  # type: str
-        self.encoding_aliases = encoding_aliases  # type: List[str]
-        self.alternative_encodings = alternative_encodings  # type: List[str]
-        self.language = language  # type: str
-        self.alphabets = alphabets  # type: List[str]
-        self.has_sig_or_bom = has_sig_or_bom  # type: bool
-        self.chaos = chaos  # type: float
-        self.coherence = coherence  # type: float
-        self.is_preferred = is_preferred  # type: bool
+    def __init__(
+        self,
+        path: str,
+        encoding: Optional[str],
+        encoding_aliases: List[str],
+        alternative_encodings: List[str],
+        language: str,
+        alphabets: List[str],
+        has_sig_or_bom: bool,
+        chaos: float,
+        coherence: float,
+        unicode_path: Optional[str],
+        is_preferred: bool,
+    ):
+        self.path: str = path
+        self.unicode_path: Optional[str] = unicode_path
+        self.encoding: Optional[str] = encoding
+        self.encoding_aliases: List[str] = encoding_aliases
+        self.alternative_encodings: List[str] = alternative_encodings
+        self.language: str = language
+        self.alphabets: List[str] = alphabets
+        self.has_sig_or_bom: bool = has_sig_or_bom
+        self.chaos: float = chaos
+        self.coherence: float = coherence
+        self.is_preferred: bool = is_preferred
 
     @property
-    def __dict__(self):
+    def __dict__(self) -> Dict[str, Any]:  # type: ignore
         return {
-            'path': self.path,
-            'encoding': self.encoding,
-            'encoding_aliases': self.encoding_aliases,
-            'alternative_encodings': self.alternative_encodings,
-            'language': self.language,
-            'alphabets': self.alphabets,
-            'has_sig_or_bom': self.has_sig_or_bom,
-            'chaos': self.chaos,
-            'coherence': self.coherence,
-            'unicode_path': self.unicode_path,
-            'is_preferred': self.is_preferred
+            "path": self.path,
+            "encoding": self.encoding,
+            "encoding_aliases": self.encoding_aliases,
+            "alternative_encodings": self.alternative_encodings,
+            "language": self.language,
+            "alphabets": self.alphabets,
+            "has_sig_or_bom": self.has_sig_or_bom,
+            "chaos": self.chaos,
+            "coherence": self.coherence,
+            "unicode_path": self.unicode_path,
+            "is_preferred": self.is_preferred,
         }
 
     def to_json(self) -> str:
-        return dumps(
-            self.__dict__,
-            ensure_ascii=True,
-            indent=4
-        )
-
-
-CharsetNormalizerMatch = CharsetMatch
+        return dumps(self.__dict__, ensure_ascii=True, indent=4)
